;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Breadth-first pathfinding algorithm                    Olivier Lecocq        22/07/2001
;                                                        (olecocq@hotmail.com)
;
; This routine computes the shortest way from one point to another point on the map. 
; The movement path is tile by tile. At each tile, you can decide between 4 possible directions
; (up,down,left,right)
;
; possible directions:
;
;                   1
;                 4< >2              
;                   3
;
; the routine can be easily modified to support more directions : e.g 8 (including diagonals)
; or 6 (hexagonals) 
;
; The method used is rather simple. It's certainly not the most effective algorithm, but it's
; guaranteed to find the shortest way. It is suited for relative small maps/distances. In theory,
; it will handle every distance, but the time/memory used to calculate the shortest path will
; increase exponentially. Another method like Dijkstra's or the A* algoritm might be suited for
; larger maps but on smaller maps those methods would require more time/memory than Breadth-first
; 
; note : the more obstacles in the map, the faster the routine performs
; 
; Method explaination:
;
; from the starting point, the program checks each of the 4 possible directions; if an empty tile
; is found, the x/y coord of that tile is added to a FIFO (first in-first out)list and the tile
; is marked with a number (in the pmap area). That number equals the amount of steps required to
; reach that point. When all directions are checked, the program moves to the next tile in the
; FIFO list and repeats the sequence until the goal tile is reached. 
; When the goal is reached, the actual path is calculated by backtracking to the starting point.
;    
; - start point
; - step 1
; - check up, right, down, left directions   <-------+
; - if direction is OK add to the list               +
; - if all directions are checked, add next step     +
; - repeat that sequence for all coord in the list --+
;
;
;                                             03
;                            02             030203          
;             01           020102         0302010203
;    00     010001       0201000102     03020100010203      ...
;             01           020102         0302010203
;                            02             030203 
;                                             03              
;
;
; map format:
; 
;   (0,0) +-----------------------+ (maxx,0)
;         +                       +
;         + (startx,starty)       +
;         +                       +
;         +                       +
;         +                       +
;         +          (endx,endy)  +
;(0,maxy) +-----------------------+ (maxx,maxy)
;
; 0 = clear path
; 1-255 = obstacle
;
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


; variables used (example a 16*8 map):

map = $9000                     ; map start address
pmap = map + mapsize            ; pathfindingmap start address 
mem1_add = $8100                ; memory 1 location address
Mem2_add = mem1_add + 320       ; memory 2 location address + **
maxx = 15                       ; x-width map 
maxy = 7                        ; y-width map 
maxsteps = 40                   ; maximum amount of steps needed to reach goal
mapsize = 128                   ; =(maxx+1)*(maxy+1)

startx = _textShadow            ; start x coord (0-maxx)
starty = startx+1               ; start y coord (0-maxy)
endx = starty+1                 ; goal x coord  (0-maxx)
endy = endx+1                   ; goal y coord  (0-maxy)

count = cury+1                  ; current amount of steps
cur_add = count+1               ; 2 bytes current mem1/mem2 address
temp_add = cur_add+2            ; 2 bytes temporary mem1/mem2 address
path_found = temp_add+2         ; path found flag    
tempx = path_found+1            ; temporary x coord
tempy = tempx+1                 ; temporary y coord
dir = tempy+1                   ; direction
path = dir+1                    ; number of bytes = maxsteps

; ** : maximum amount of bytes needed = 2 * (4 * maxsteps). 
;      this is the amount of possible paths at a maxsteps distance from the starting point
;      when not a single obstacle or map border is reached. In most cases, you won't need that 
;      amount of memory space. As a thumb rule, 4 * maxsteps will do. 
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                
find_path:
    ld a,0
    ld (path_found),a       ; 0= path not found;1= path found
    call Clear_pmap         ; clear the temporary map
    call delete_mem1        
    call delete_mem2        ; clear the memory areas used for this routine
    ld a,1                  ; 1= first step
    ld (count),a            ; save it
    ld hl,mem1_add          
    ld (temp_add),hl        ; save the address used to save the coord of the tiles which will be found               
    ld (hl),$ff             ; $FF means end of list
    ld a,(startx)
    ld b,a
    ld a,(starty)
    ld c,a                  ; get start coord       
    call calc_map_pos       ; calculate the map address for that tile
    ld de,mapsize
    add hl,de               ; address of starin point on pmap
    ld (hl),100             ; 100= code on pmap for starting point
    ld hl,mem2_add
    ld a,(startx)
    ld (hl),a
    ld a,(starty)
    inc hl
    ld (hl),a               ; save start coord in mem2_add
    inc hl        
    ld (hl),$ff             ; end of list (only 1 coord = start point)
    ld hl,mem2_add
    ld (cur_add),hl         ; load the starting point in current address
    
Next_step:                  ; start loop
    ld hl,(cur_add)         ; get the current tile coord
    ld a,(hl)
    cp $ff                  ; end of list?
    jp z,step_done          ; if end of list reached, proceed with the next step
    ld (tempx),a            
    inc hl
    ld a,(hl)
    ld (tempy),a            ; save the current x/y coord      
    cp 0                    ; check if upper border
    jp z, skip_up           ; upper border reached    
    ld c,a
    ld a,(tempx)
    ld b,a
    dec c                   ; check up 
    call Check_dir    
skip_up:    
    ld a,(tempx)
    cp maxx
    jp z, skip_right        ; right border reached
    ld b,a
    inc b                   ; check right
    ld a,(tempy)
    ld c,a
    call Check_dir
skip_right:
    ld a,(tempy)
    cp maxy
    jp z,Skip_down          ; lower border reached
    ld c,a
    inc c                   ; check lower
    ld a,(tempx)
    ld b,a
    call Check_dir
skip_down:
    ld a,(tempx)
    cp 0
    jp z,skip_done          ; left border reached
    ld b,a
    dec b                   ; check left
    ld a,(tempy)
    ld c,a
    call Check_dir
skip_done:
    ld a,(path_found)
    cp 1
    jp z,calc_path           ; goal is reached, now calculate the actual path  
    ld hl,(cur_add)
    inc hl
    inc hl
    ld (cur_add),hl
    jp Next_step             ; move pointer 2 places forward to next tile coord
Step_done:    
    ld a,(count)             ; get amount of steps so far
    cp maxsteps              ; max steps reached, goal not found
    jp z,Mainloop            ; exit pathfinding routine, failed to find path
    inc a                    ; increase steps
    ld (count),a             ; save amount of steps
    rra
    jp c,odd_count           ; odd steps   
    jp even_count            ; even steps
                 
Even_count:                  ; swap memories
    call delete_mem2                  
    ld hl,mem1_add
    ld (cur_add),hl
    ld hl,mem2_add
    ld (temp_add),hl
    ld (hl),$FF              ; stopcode
    jp next_step
Odd_count:                   ; swap memories   
    call delete_mem1
    ld hl,mem2_add
    ld (cur_add),hl
    ld hl,mem1_add
    ld (temp_add),hl 
    ld (hl),$ff             ;stopcode
    jp next_step

Check_dir:                   
    push bc                 ; bc= xy coord of tile to check on map
    call Calc_map_pos       ; calculate map address
    pop bc
    ld de,mapsize
    add hl,de               ; pmap address
    ld a,(hl)               ; get pmap byte
    cp 0
    ret nz                  ; tile already checked    
    sbc hl,de               ; map address
    ld a,(hl)               ; get map byte
    cp 0
    call z,Add_free_node    ; that tile is a possible direction so save it in list
    cp 6                    ; 6=goal code on map
    call z,goal_reached     ; goal is reached
    ret

Add_free_node:
    ld de,mapsize
    add hl,de               ; pmap address
    ld a,(count)            ; get steps
    ld (hl),a               ; save distance on pmap
    ld hl,(temp_add)        ; get the memory address to save the coord in list
    ld (hl),b               ; save x coord
    inc hl
    ld (hl),c               ; save y coord
    inc hl
    ld (hl),$FF             ; stopcode
    ld (temp_add),hl        ; save next mem slot 
    xor a
    ret
    
Calc_path:                  ; 2nd part of the routine, backtracking from goal
    ld a,(count)            ; load the amount of steps needed to reach destination
    ld hl,path              ; load the memory address wher you will save the path
    ld b,0
    ld c,a
    add hl,bc               ; you need (steps) bytes to save the path (saves the direction at each tile)
    ld (cur_add),hl         
    ld a,(endx)
    ld (tempx),a
    ld a,(endy)
    ld (tempy),a            ; save the end point
    ld a,(count)
    cp 1
    jp nz,Calc_path_loop    ; more than 1 step neede to reach goal
    ld a,101
    ld (count),a            ; only step so you only have to find the direction of the start point with code 100
Calc_path_loop:    
    ld a,3                  ; check first direction (above the current tile)
    ld (dir),a              ; saves the direction
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a    
    ld a,c
    cp 0
    jp z,skip_dir1          ; upper border is reached
    dec a
    ld c,a                  ; y = y - 1
    call check_tile
skip_dir1:
    ld a,4                  ; check tile to the right of current tile
    ld (dir),a              ; save the direction
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,b
    cp maxx         
    jp z,skip_dir2          ; right border map
    inc a
    ld b,a                  ; x = x + 1
    call check_tile
skip_dir2:
    ld a,1                  ; check tile under the current tile
    ld (dir),a              ; saves direction
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,c
    cp maxy                 
    jp z,skip_dir3          ; lower map border
    inc a
    ld c,a                  ; y = y + 1
    call check_tile
skip_dir3:
    ld a,2                  ; check tile to the left
    ld (dir),a              ; saves direction
    ld a,(tempx)
    ld b,a
    ld a,(tempy)
    ld c,a
    ld a,b
    cp 0
    jp z,skip_dir4          ; left map border
    dec a
    ld b,a                  ; x = x - 1
    call check_tile
skip_dir4:
    ld a,(count)
    cp 100                  ; 100 = start code
    jp nz,calc_path_loop    ; not yet found, loop
    jp Mainloop             ; path found, end the pathfinding routine    

Check_tile:
    push bc                 ; save xy coord    
    call calc_map_pos
    pop bc
    ld de,mapsize
    add hl,de               ; pmap address
    ld a,(count)    
    dec a                   ; decrease steps
    cp (hl)                 ; does that tile have a number that is one less then the current tile?.
    ret nz                  ; nope, return to the loop to check other direction
    ld a,(dir)              ; tile found
    ld hl,(cur_add)         ; saves the direction
    ld (hl),a
    dec hl
    ld (cur_add),hl         ; new address to save direction
    ld a,(count)
    cp 2
    jp z,almost_there       ; only one more step to go
    dec a
    ld (count),a            ; decrease steps
    ret
almost_there:
    ld a,101                ; 101-1=100 = start point 
    ld (count),a
    ret
                        
Goal_reached:
    ld a,1
    ld (path_found),a
    ret   
                
delete_mem1:
    ld hl,mem1_add
    ld b,320            ; 320 bytes reserved depends of maxsteps (see above)
    call _ldhlz
    ret
    
delete_mem2:
    ld hl,mem2_add
    ld b,320
    call _ldhlz
    ret

Calc_map_pos:           ; input b=x coord ; c=y coord ; output hl=map address  
    push bc
    ld a,c
    cp 0
    jp z,Calc_y_done
    ld a,(maxx)
    ld d,0
    ld e,a
    ld b,c
    ld hl,map
New_Y_line:
    add hl,de        
    djnz New_Y_line   
Calc_y_done:
    pop bc
    ld c,b
    ld b,0
    add hl,bc
    ret
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                                                                               
 